from hdmf.testing import TestCase

from nwbinspector import Importance
from nwbinspector.utils import format_byte_size, check_regular_series, is_dict_in_string


def test_format_byte_size():
    assert format_byte_size(byte_size=12345) == "12.35KB"


def test_format_byte_size_in_binary():
    assert format_byte_size(byte_size=12345, units="binary") == "12.06KiB"


class TestFormatByteException(TestCase):
    def test_format_byte_size_units_exception(self):
        with self.assertRaisesWith(
            exc_type=ValueError,
            exc_msg="'units' argument must be either 'SI' (for orders of 1000) or 'binary' (for orders of 1024).",
        ):
            format_byte_size(byte_size=12345, units="test")


def test_check_regular_series():
    assert check_regular_series(series=[1, 2, 3])
    assert not check_regular_series(series=[1, 2, 4])


def test_is_dict_in_string_false_1():
    string = "not a dict"
    assert is_dict_in_string(string=string) is False


def test_is_dict_in_string_false_2():
    string = "not a dict, {but also fancy format text!}"
    assert is_dict_in_string(string=string) is False


def test_is_dict_in_string_false_3():
    string = "[not] a dict, {[but] also} fancy format text!"
    assert is_dict_in_string(string=string) is False


def test_is_dict_in_string_true_1():
    string = str(dict(a=1))
    assert is_dict_in_string(string=string) is True


def test_is_dict_in_string_true_2():
    string = str([dict(a=1), dict(b=2)])
    assert is_dict_in_string(string=string) is True


def test_is_dict_in_string_true_3():
    string = str(dict(a=dict(b=2)))
    assert is_dict_in_string(string=string) is True


def test_is_dict_in_string_true_4():
    string = "some text: {'then': 'a dict'}"
    assert is_dict_in_string(string=string) is True


def test_is_dict_in_string_true_5():
    string = "{'then': 'a dict'} more text"
    assert is_dict_in_string(string=string) is True


def test_is_dict_in_string_true_6():
    """Not a JSON encodable object."""
    string = str({1.2: Importance.CRITICAL})
    assert is_dict_in_string(string=string) is True


def test_is_dict_in_string_true_7():
    """
    A more aggressive demonstration of the general dictionary regex.

    But it is technically possible to achieve via `str({custom_object_1: custom_object_2})` if 'custom_object_1' is
    hashable and both custom objects have manual `__repr__` that do not include apostrophe's within their return.

    Example
    -------
    from dataclasses import dataclass

    @dataclass
    class Test():
        prop = 1

        def __repr__(self):
            return "This is a test"

    str({1: Test()})
    """
    string = "example, {this is not a dict: but it sure looks like one}!"
    assert is_dict_in_string(string=string) is True



"""Check functions specific to optical electrophysiology neurodata types."""
from pynwb.ophys import (
    RoiResponseSeries,
    PlaneSegmentation,
    OpticalChannel,
    ImagingPlane,
)

from hdmf.utils import get_data_shape

from ..register_checks import register_check, Importance, InspectorMessage

MIN_LAMBDA = 10.0  # trigger warnings for wavelength values less than this value


@register_check(importance=Importance.CRITICAL, neurodata_type=RoiResponseSeries)
def check_roi_response_series_dims(roi_response_series: RoiResponseSeries):
    """Check the dimensions of an ROI series to ensure the time axis is the correct dimension."""
    data = roi_response_series.data
    rois = roi_response_series.rois

    data_shape = get_data_shape(data, strict_no_data_load=True)

    if data_shape and len(data_shape) == 2 and data_shape[1] != len(rois.data):
        if data_shape[0] == len(rois.data):
            return InspectorMessage(
                message=(
                    "The second dimension of data does not match the length of rois, "
                    "but instead the first does. Data is oriented incorrectly and should be transposed."
                )
            )
        return InspectorMessage(
            message="The second dimension of data does not match the length of rois. Your data may be transposed."
        )


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=RoiResponseSeries)
def check_roi_response_series_link_to_plane_segmentation(
    roi_response_series: RoiResponseSeries,
):
    """Check that each ROI response series links to a plane segmentation."""
    if not isinstance(roi_response_series.rois.table, PlaneSegmentation):
        return InspectorMessage(message="rois field does not point to a PlaneSegmentation table.")


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=OpticalChannel)
def check_emission_lambda_in_nm(optical_channel: OpticalChannel):
    """Check that emission lambda is in feasible range for unit nanometers."""
    if optical_channel.emission_lambda < MIN_LAMBDA:
        return InspectorMessage(f"emission lambda of {optical_channel.emission_lambda} should be in units of nm.")


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=ImagingPlane)
def check_excitation_lambda_in_nm(imaging_plane: ImagingPlane):
    """Check that emission lambda is in feasible range for unit nanometers."""
    if imaging_plane.excitation_lambda < MIN_LAMBDA:
        return InspectorMessage(f"excitation lambda of {imaging_plane.excitation_lambda} should be in units of nm.")



from unittest import TestCase
from tempfile import mkdtemp
from shutil import rmtree
from pathlib import Path

import numpy as np
from pynwb import NWBHDF5IO
from pynwb.image import ImageSeries

from nwbinspector import (
    InspectorMessage,
    Importance,
    check_image_series_external_file_valid,
    check_image_series_external_file_relative,
)
from nwbinspector.tools import make_minimal_nwbfile


class TestExternalFileValid(TestCase):
    def setUp(self):
        self.tempdir = Path(mkdtemp())
        self.tempfile = self.tempdir / "tempfile.mov"
        self.nested_tempdir_2 = self.tempdir / "nested_dir"
        self.nested_tempdir_2.mkdir(parents=True)
        self.tempfile2 = self.nested_tempdir_2 / "tempfile2.avi"
        for file in [self.tempfile, self.tempfile2]:
            with open(file=file, mode="w") as fp:
                fp.write("Not a movie file, but at least it exists.")
        self.nwbfile = make_minimal_nwbfile()
        self.nwbfile.add_acquisition(
            ImageSeries(
                name="TestImageSeries",
                rate=1.0,
                external_file=[
                    "/".join([".", self.tempfile.name]),
                    "/".join([".", self.tempfile2.parent.stem, self.tempfile2.name]),
                ],
            )
        )
        self.nwbfile.add_acquisition(
            ImageSeries(
                name="TestImageSeriesBad1",
                rate=1.0,
                external_file=["madeup_file.mp4"],
            )
        )
        self.absolute_file_path = str(Path("madeup_file.mp4").absolute())
        self.nwbfile.add_acquisition(
            ImageSeries(name="TestImageSeriesBad2", rate=1.0, external_file=[self.absolute_file_path])
        )
        image_module = self.nwbfile.create_processing_module(name="behavior", description="testing imageseries")
        image_module.add(ImageSeries(name="TestImageSeries2", rate=1.0, external_file=[self.tempfile, self.tempfile2]))
        with NWBHDF5IO(path=self.tempdir / "tempnwbfile.nwb", mode="w") as io:
            io.write(self.nwbfile)

    def tearDown(self):
        rmtree(self.tempdir)

    def test_check_image_series_external_file_valid_pass(self):
        with NWBHDF5IO(path=self.tempdir / "tempnwbfile.nwb", mode="r") as io:
            nwbfile = io.read()
            assert check_image_series_external_file_valid(image_series=nwbfile.acquisition["TestImageSeries"]) is None

    def test_check_image_series_external_file_valid_bytestring_pass(self):
        """Can't call the io.write() step in setUp as that decodes the bytes with our version of h5py."""
        nwbfile = make_minimal_nwbfile()
        nwbfile.add_acquisition(
            ImageSeries(
                name="TestImageSeries",
                rate=1.0,
                external_file=[bytes("/".join([".", self.tempfile.name]), "utf-8")],
            )
        )
        assert check_image_series_external_file_relative(image_series=nwbfile.acquisition["TestImageSeries"]) is None

    def test_check_image_series_external_file_valid(self):
        with NWBHDF5IO(path=self.tempdir / "tempnwbfile.nwb", mode="r") as io:
            nwbfile = io.read()
            image_series = nwbfile.acquisition["TestImageSeriesBad1"]
            assert check_image_series_external_file_valid(image_series=image_series)[0] == InspectorMessage(
                message=(
                    "The external file 'madeup_file.mp4' does not exist. Please confirm the relative location to the"
                    " NWBFile."
                ),
                importance=Importance.CRITICAL,
                check_function_name="check_image_series_external_file_valid",
                object_type="ImageSeries",
                object_name="TestImageSeriesBad1",
                location="/acquisition/TestImageSeriesBad1",
            )

    def test_check_image_series_external_file_relative_pass(self):
        with NWBHDF5IO(path=self.tempdir / "tempnwbfile.nwb", mode="r") as io:
            nwbfile = io.read()
            assert (
                check_image_series_external_file_relative(image_series=nwbfile.acquisition["TestImageSeries"]) is None
            )

    def test_check_image_series_external_file_relative_bytestring_pass(self):
        """Can't call the io.write() step in setUp as that decodes the bytes with our version of h5py."""
        image_series = ImageSeries(
            name="TestImageSeries",
            rate=1.0,
            external_file=[bytes("/".join([".", self.tempfile.name]), "utf-8")],
        )
        assert check_image_series_external_file_relative(image_series=image_series) is None

    def test_check_image_series_external_file_relative(self):
        with NWBHDF5IO(path=self.tempdir / "tempnwbfile.nwb", mode="r") as io:
            nwbfile = io.read()
            image_series = nwbfile.acquisition["TestImageSeriesBad2"]
            assert check_image_series_external_file_relative(image_series=image_series)[0] == InspectorMessage(
                message=(
                    f"The external file '{self.absolute_file_path}' is not a relative path. "
                    "Please adjust the absolute path to be relative to the location of the NWBFile."
                ),
                importance=Importance.BEST_PRACTICE_VIOLATION,
                check_function_name="check_image_series_external_file_relative",
                object_type="ImageSeries",
                object_name="TestImageSeriesBad2",
                location="/acquisition/TestImageSeriesBad2",
            )


def test_check_image_series_external_file_valid_pass_non_external():
    image_series = ImageSeries(name="TestImageSeries", rate=1.0, data=np.zeros(shape=(3, 3, 3, 3)), unit="TestUnit")
    assert check_image_series_external_file_valid(image_series=image_series) is None



from pynwb.behavior import SpatialSeries
import numpy as np

from nwbinspector import InspectorMessage, Importance, check_spatial_series_dims


def test_check_spatial_series_dims():

    spatial_series = SpatialSeries(
        name="SpatialSeries",
        description="description",
        data=np.ones((10, 4)),
        rate=3.0,
        reference_frame="reference_frame",
    )
    assert check_spatial_series_dims(spatial_series) == InspectorMessage(
        message="SpatialSeries should have 1 column (x), 2 columns (x, y), or 3 columns (x, y, z).",
        importance=Importance.CRITICAL,
        check_function_name="check_spatial_series_dims",
        object_type="SpatialSeries",
        object_name="SpatialSeries",
        location="/",
    )


def test_pass_check_spatial_series_dims():

    spatial_series = SpatialSeries(
        name="SpatialSeries",
        description="description",
        data=np.ones((10, 3)),
        rate=3.0,
        reference_frame="reference_frame",
    )

    assert check_spatial_series_dims(spatial_series) is None


def test_pass_check_spatial_series_dims_1d():

    spatial_series = SpatialSeries(
        name="SpatialSeries",
        description="description",
        data=np.ones((10,)),
        rate=3.0,
        reference_frame="reference_frame",
    )

    assert check_spatial_series_dims(spatial_series) is None



"""Checks for types belonging to the pynwb.behavior module."""
from pynwb.behavior import SpatialSeries

from ..register_checks import register_check, Importance, InspectorMessage


@register_check(importance=Importance.CRITICAL, neurodata_type=SpatialSeries)
def check_spatial_series_dims(spatial_series: SpatialSeries):
    """Check if a SpatialSeries has the correct dimensions."""
    if len(spatial_series.data.shape) > 1 and spatial_series.data.shape[1] > 3:
        return InspectorMessage(
            message="SpatialSeries should have 1 column (x), 2 columns (x, y), or 3 columns (x, y, z)."
        )
